def permutation_sample(data_1, data_2):
    """
    Generate a permutation sample from two data sets. Specifically,
    concatenate `data_1` and `data_2`, scramble the order of the
    concatenated array, and then return the first len(data_1) entries
    and the last len(data_2) entries.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    out_1 : ndarray, same shape as `data_1`
        Permutation sample corresponding to `data_1`.
    out_2 : ndarray, same shape as `data_2`
        Permutation sample corresponding to `data_2`.
    """
    data_1 = utils._convert_data(data_1)
    data_2 = utils._convert_data(data_2)

    return _permutation_sample(data_1, data_2)


@numba.jit(nopython=True)
def _permutation_sample(data_1, data_2):
    """
    Generate a permutation sample from two data sets. Specifically,
    concatenate `data_1` and `data_2`, scramble the order of the
    concatenated array, and then return the first len(data_1) entries
    and the last len(data_2) entries.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    out_1 : ndarray, same shape as `data_1`
        Permutation sample corresponding to `data_1`.
    out_2 : ndarray, same shape as `data_2`
        Permutation sample corresponding to `data_2`.
    """
    x = np.concatenate((data_1, data_2))
    np.random.shuffle(x)
    return x[: len(data_1)], x[len(data_1) :]



import numpy as np
import pandas as pd


def to_dataframe(
    fit, pars=None, permuted=False, dtypes=None, inc_warmup=False, diagnostics=True
):
    """
    Convert the results of a fit to a DataFrame.

    Parameters
    ----------
    fit : StanFit4Model instance
        The output of fitting a Stan model.
    pars : str or list of strs
        The parameters to extract into the outputted data frame.
    permuted : bool, default False
        If True, the chains are combined and the samples are permuted.
    dypes : dict
        Each entry key, value pair is param_name, data dtype. If None,
        all data types are floats. Be careful: results are cast into
        whatever you specify.
    inc_warmup : bool, default False
        If True, include warmup samples.
    diagnostics : bool, default True
        If True, also include diagnostic information about the samples.
        Note that if no Monte Carlo sampling is done (e.g., if the
        Fixed_param algorithm is used).

    Returns
    -------
    output : DataFrame
        A Pandas DataFrame containing the samples. Each row consists of
        a single sample, including the parameters and diagnostics if
        `diagnostics` is True.

    Notes
    -----
    .. This functionality is present in PyStan >= 2.18. Due to
       compilation issues with higher versions of PyStan, we include
       this functionality here. The output of this function is meant to
       match that of fit.to_dataframe() of a StanFit4Model instance
       from PyStan 2.18 and above. Because of PyStan's GPL license,
       I have re-written this functionality here to enable permissive
       licensing of this module.
    """
    if permuted and inc_warmup:
        raise RuntimeError("If `permuted` is True, `inc_warmup` must be False.")

    if permuted and diagnostics:
        raise RuntimeError("Diagnostics are not available when `permuted` is True.")

    if dtypes is not None and not permuted and pars is None:
        raise RuntimeError(
            "`dtypes` cannot be specified when `permuted`"
            + " is False and `pars` is None."
        )

    try:
        return fit.to_dataframe(
            pars=pars,
            permuted=permuted,
            dtypes=dtypes,
            inc_warmup=inc_warmup,
            diagnostics=diagnostics,
        )
    except:
        pass

    # Diagnostics to pull out
    diags = [
        "diverging__",
        "energy__",
        "treedepth__",
        "accept_stat__",
        "stepsize__",
        "n_leapfrog__",
    ]

    # Build parameters if not supplied
    if pars is None:
        pars = tuple(fit.flatnames + ["lp__"])
    if type(pars) not in [list, tuple, pd.core.indexes.base.Index]:
        raise RuntimeError("`pars` must be list or tuple or pandas index.")
    if "lp__" not in pars:
        pars = tuple(list(pars) + ["lp__"])

    # Build dtypes if not supplied
    if dtypes is None:
        dtypes = {par: float for par in pars}
    else:
        dtype["lp__"] = float

    # Make sure dtypes supplied for every parameter
    for par in pars:
        if par not in dtypes:
            raise RuntimeError(f"'{par}' not in `dtypes`.")

    # Retrieve samples
    samples = fit.extract(
        pars=pars, permuted=permuted, dtypes=dtypes, inc_warmup=inc_warmup
    )
    n_chains = len(fit.stan_args)
    thin = fit.stan_args[0]["thin"]
    n_iters = fit.stan_args[0]["iter"] // thin
    n_warmup = fit.stan_args[0]["warmup"] // thin
    n_samples = n_iters - n_warmup

    # Dimensions of parameters
    dim_dict = {par: dim for par, dim in zip(fit.model_pars, fit.par_dims)}
    dim_dict["lp__"] = []

    if inc_warmup:
        n = (n_warmup + n_samples) * n_chains
        warmup = np.concatenate(
            [[1] * n_warmup + [0] * n_samples for _ in range(n_chains)]
        ).astype(int)
        chain = np.concatenate(
            [[i + 1] * (n_warmup + n_samples) for i in range(n_chains)]
        ).astype(int)
        chain_idx = np.concatenate(
            [np.arange(1, n_warmup + n_samples + 1) for _ in range(n_chains)]
        ).astype(int)
    else:
        n = n_samples * n_chains
        warmup = np.array([0] * n, dtype=int)
        chain = np.concatenate([[i + 1] * n_samples for i in range(n_chains)]).astype(
            int
        )
        chain_idx = np.concatenate(
            [np.arange(1, n_samples + 1) for _ in range(n_chains)]
        ).astype(int)

    if permuted:
        df = pd.DataFrame()
    else:
        df = pd.DataFrame(data=dict(chain=chain, chain_idx=chain_idx, warmup=warmup))

    if diagnostics:
        sampler_params = fit.get_sampler_params(inc_warmup=inc_warmup)
        diag_vals = list(sampler_params[0].keys())

        # If they are standard, do same order as PyStan 2.18 to_dataframe
        if (
            len(diag_vals) == len(diags)
            and sum([val not in diags for val in diag_vals]) == 0
        ):
            diag_vals = diags
        for diag in diag_vals:
            df[diag] = np.concatenate(
                [sampler_params[i][diag] for i in range(n_chains)]
            )
            if diag in ["treedepth__", "n_leapfrog__"]:
                df[diag] = df[diag].astype(int)

    if isinstance(samples, np.ndarray):
        for k, par in enumerate(pars):
            try:
                indices = re.search("\[(\d+)(,\d+)*\]", par).group()
                base_name = re.split("\[(\d+)(,\d+)*\]", par, maxsplit=1)[0]
                col = base_name + re.sub(
                    "\d+", lambda x: str(int(x.group()) + 1), indices
                )
            except AttributeError:
                col = par

            df[col] = samples[:, :, k].flatten(order="F")
    else:
        for par in pars:
            if len(dim_dict[par]) == 0:
                df[par] = samples[par].flatten(order="F")
            else:
                for inds in itertools.product(*[range(dim) for dim in dim_dict[par]]):
                    col = (
                        par + "[" + ",".join([str(ind + 1) for ind in inds[::-1]]) + "]"
                    )

                    if permuted:
                        array_slice = tuple([slice(n), *inds[::-1]])
                    else:
                        array_slice = tuple([slice(n), slice(n), *inds[::-1]])

                    df[col] = samples[par][array_slice].flatten(order="F")

    return df


def to_arviz(fit, log_likelihood=None, logging_disable_level=logging.ERROR):
    """Convert a StanFit4Model to an ArviZ data set.

    Parameters
    ----------
    fit : StanFit4Model instance
        Samples from a Stan calculation.
    log_likelihood : str
        Name of the variable containing the log likelihood.
    logging_disable_level : str, default logging.ERROR
        Logging is disabled below this level. This helps deal with
        annoying warning from PyStan when doing extractions when dtype
        is not specified and permuted is False.

    Returns
    -------
    output : ArviZ data object.

    Notes
    -----
    .. This function is only necessary because of problems ArviZ has
       with PyStan < 2.18. You should just directly use
       `arviz.from_pystan()` if you have PyStan >= 2.18. Note that this
       function does not allow other kwargs besides `log_likelihood`
       that are available in `arviz.from_pystan()`. The main purpose
       of this function is to get minimal information necessary for
       doing PSIS-LOO and WAIC calculations using ArviZ.
    """
    with _disable_logging(level=logging_disable_level):
        az_data = az.from_pystan(fit=fit, log_likelihood=log_likelihood)

    if pystan.__version__ < "2.18":
        if log_likelihood is not None:
            # Get the log likelihood
            log_lik = np.swapaxes(
                extract_par(
                    fit, log_likelihood, logging_disable_level=logging_disable_level
                ),
                0,
                1,
            )

            # dims for xarray
            dims = ["chain", "draw", "log_likelihood_dim_0"]

            # Properly do the log likelihood
            az_data.sample_stats["log_likelihood"] = (dims, log_lik)

        # Get the lp__
        with _disable_logging(level=logging_disable_level):
            lp = np.swapaxes(fit.extract(permuted=False)[:, :, -1], 0, 1)
        az_data.sample_stats["lp"] = (["chain", "draw"], lp)

    return az_data

def waic(
    fit, log_likelihood=None, pointwise=False, logging_disable_level=logging.ERROR
):
    """Compute the WAIC using ArviZ.

    Parameters
    ----------
    fit : StanFit4Model instance
        Samples from a Stan calculation.
    log_likelihood : str
        Name of the variable containing the log likelihood.
    pointwise : bool, default False
        If True, also return point-wise WAIC.
    logging_disable_level : str, default logging.ERROR
        Logging is disabled below this level. This helps deal with
        annoying warning from PyStan when doing extractions when dtype
        is not specified and permuted is False.

    Returns
    -------
    output : Pandas data frame
        Pandas DataFrame with columns:
          waic: widely available information criterion
          waic_se: standard error of waic
          p_waic: effective number parameters
          var_warn: 1 if posterior variance of the log predictive
            densities exceeds 0.4
          waic_i: and array of the pointwise predictive accuracy, only
            if `pointwise` True
    """
    if log_likelihood is None:
        raise RuntimeError("Must supply `log_likelihood`.")

    az_data = to_arviz(
        fit, log_likelihood=log_likelihood, logging_disable_level=logging_disable_level
    )

    return az.waic(az_data, pointwise=pointwise)


def loo(
    fit,
    log_likelihood=None,
    pointwise=False,
    reff=None,
    logging_disable_level=logging.ERROR,
):
    """Compute the PSIS-LOO.

    Parameters
    ----------
    fit : StanFit4Model instance
        Samples from a Stan calculation.
    log_likelihood : str
        Name of the variable containing the log likelihood.
    pointwise : bool, default False
        If True, also return point-wise predictive accuracy.
    reff : float, optional
        Relative MCMC efficiency, `effective_n / n` i.e. number of
        effective samples divided by the number of actual samples.
        Computed from trace by default.
    logging_disable_level : str, default logging.ERROR
        Logging is disabled below this level. This helps deal with
        annoying warning from PyStan when doing extractions when dtype
        is not specified and permuted is False.

    Returns
    -------
    output : Pandas data frame
        Pandas DataFrame with columns:
          loo: approximated Leave-one-out cross-validation
          loo_se: standard error of loo
          p_loo: effective number of parameters
          shape_warn: 1 if the estimated shape parameter of Pareto
            distribution is greater than 0.7 for one or more samples.
          loo_i: array of pointwise predictive accuracy, only if
            `pointwise` True
    """
    if log_likelihood is None:
        raise RuntimeError("Must supply `log_likelihood`.")

    az_data = to_arviz(
        fit, log_likelihood=log_likelihood, logging_disable_level=logging_disable_level
    )

    return az.loo(az_data, pointwise=pointwise, reff=reff)


def compare(
    fit_dict, log_likelihood=None, logging_disable_level=logging.ERROR, **kwargs
):
    """Compare models.

    Parameters
    ----------
    fit_dict : StanFit4Model instance
        A dictionary where each key it a name attached to a model and
        the value is a StanFit4Model instance that has the posterior
        samples.
    log_likelihood : str
        Name of the variable containing the log likelihood.
    logging_disable_level : str, default logging.ERROR
        Logging is disabled below this level. This helps deal with
        annoying warning from PyStan when doing extractions when dtype
        is not specified and permuted is False.

    Returns
    -------
    output : Pandas data frame
        Pandas DataFrame with columns:
          IC : Information Criteria (WAIC or LOO).
          pIC : Estimated effective number of parameters.
          dIC : Relative difference between each IC (WAIC or LOO) and
            the lowest IC (WAIC or LOO). It is always 0 for the
            top-ranked model.
          weight: Relative weight for each model.
            This can be loosely interpreted as the probability of each
            model (among the compared model) given the data. By default
            the uncertainty in the weights estimation is considered using
            Bayesian bootstrap.
          SE : Standard error of the IC estimate. If `method` is
            BB-pseudo-BMA, these values are estimated using Bayesian
            bootstrap.
          dSE : Standard error of the difference in IC between each
            model and the top-ranked model. It is always 0 for the
            top-ranked model.
          warning : A value of 1 indicates that the computation of the
            IC may not be reliable. This could be indication of WAIC/LOO
            starting to fail; see http://arxiv.org/abs/1507.04544.

    Notes
    -----
    .. All kwargs are passed into arviz.stats.compare(). These kwargs
       are given in the ArviZ documentation. Importantly, use
       `ic='waic'` or `ic='loo'` to respectively use WAIC or LOO as the
       information criterion. WAIC is the default. Use
       `method='stacking'` for stacking to compute weights (default) and
       `method='BB-pseudo-BMA'` to use pseudo-Bayesian model averaging
       with Akaike-type weights.

    """
    if log_likelihood is None:
        raise RuntimeError("Must supply `log_likelihood`.")

    arviz_dict = {}
    for key in fit_dict:
        arviz_dict[key] = to_arviz(
            fit_dict[key],
            log_likelihood=log_likelihood,
            logging_disable_level=logging_disable_level,
        )

    return az.compare(arviz_dict, **kwargs)


def pickle_dump_samples(fit=None, model=None, pkl_file=None):
    """Dump samples into a pickle file.

    Parameters
    ----------
    fit : StanFit4Model instance
        The output of fitting a Stan model.
    model : StanModel instance
        StanModel instance used to get `fit`.
    pkl_file : str
        Name of pickle file to create and dump result.

    Returns
    -------
    None
    """
    if None in [fit, model, pkl_file]:
        raise RuntimeError("`fit`, `model`, and `pkl_file` must all be specified.")

    if os.path.isfile(pkl_file):
        raise RuntimeError(f"File {pkl_file} already exists.")

    with open(pkl_file, "wb") as f:
        pickle.dump({"model": model, "fit": fit}, f, protocol=-1)


def pickle_load_samples(pkl_file):
    """Load samples out of a pickle file.

    Parameters
    ----------
    fit : StanFit4Model instance
        The output of fitting a Stan model.
    model : StanModel instance
        StanModel instance used to get `fit`.
    pkl_file : str
        Name of pickle file to create and dump result.

    Returns
    -------
    None
    """
    with open(pkl_file, "rb") as f:
        data_dict = pickle.load(f)

    return data_dict["fit"], data_dict["model"]


def extract_array(samples, name):
    """Extract an array values from a DataFrame containing samples.

    Parameters
    ----------
    samples : StanFit4Model instance or Pandas DataFrame
        Samples from a Stan calculation that contains an array for
        extraction.
    name : str
        The name of the array to extract. For example, if `name` is
        'my_array', and the array is one-dimensional, entries with
        variable names 'my_array[1]', 'my_array[2]', ... are extracted.

    Returns
    -------
    output : Pandas DataFrame
        A tidy DataFrame with the extracted array. For a 1-D array, the
        DataFrame has columns:
            ['chain', 'chain_idx', 'warmup', 'index_1', 'name']
        For a 2D array, there is an additional column named 'index_2'. A
        3D array has a column 'index_3' and so on.
    """
    df = _fit_to_df(samples, diagnostics=False)

    regex_name = name
    for char in "[\^$.|?*+(){}":
        regex_name = regex_name.replace(char, "\\" + char)

    # Extract columns that match the name
    sub_df = df.filter(regex=regex_name + "\[(\d+)(,\d+)*\]")

    if len(sub_df.columns) == 0:
        raise RuntimeError(
            "column '{}' is either absent or scalar-valued.".format(name)
        )

    n_entries = len(sub_df.columns)
    n = len(sub_df)

    df_out = pd.DataFrame(data={name: sub_df.values.flatten(order="F")})
    for col in ["chain", "chain_idx", "warmup"]:
        if col in df:
            df_out[col] = np.concatenate([df[col].values] * n_entries)

    indices = [
        re.search("\[(\d+)(,\d+)*\]", col).group()[1:-1].split(",")
        for col in sub_df.columns
    ]
    indices = np.vstack([np.array([[int(i) for i in ind]] * n) for ind in indices])
    ind_df = pd.DataFrame(
        columns=["index_{0:d}".format(i) for i in range(1, indices.shape[1] + 1)],
        data=indices,
    )

    return pd.concat([ind_df, df_out], axis=1)


def extract_par(
    fit, par, permuted=False, inc_warmup=False, logging_disable_level=logging.ERROR
):
    """Extract samples of a parameter out of a StanFit4Model.

    Parameters
    ----------
    fit : StanFit4Model instance
        Samples from a Stan calculation.
    par : str
        The name of the variable to extract. For example, if `name` is
        'my_array', and the array is one-dimensional, entries with
        variable names 'my_array[1]', 'my_array[2]', ... are extracted.
    permuted : bool, default False
        If True, the chains are combined and the samples are permuted.
    inc_warmup : bool, default False
        If True, include warmup samples.
    logging_disable_level : str, default logging.ERROR
        Logging is disabled below this level. This helps deal with
        annoying warning from PyStan when doing extractions when dtype
        is not specified and permuted is False.

    Returns
    -------
    output : Numpy array
        Numpy array with the samples. The first dimension of
        the array is for the iterations; the second for the number of chains;
        the third for the parameters. Vectors and arrays are expanded to one
        parameter (a scalar) per cell, with names indicating the third dimension.

    Notes
    -----
    .. PyStan 2.17.1.0 has a bug where the `extract()` functionality of
       a StanFit4Model instance does not work as advertised. This
       accommodates that functionality.
    """
    if permuted:
        return fit.extract(pars=par, permuted=True, dtypes=None, inc_warmup=inc_warmup)[
            par
        ]

    if pystan.__version__ >= "2.18":
        with _disable_logging(level=logging_disable_level):
            return fit.extract(
                pars=par, permuted=False, dtypes=None, inc_warmup=inc_warmup
            )

    with _disable_logging(level=logging_disable_level):
        data = fit.extract(permuted=False, inc_warmup=inc_warmup, dtypes=None)

    inds = []
    for k, par_name in enumerate(fit.flatnames):
        try:
            indices = re.search("\[(\d+)(,\d+)*\]", par_name).group()
            base_name = re.split("\[(\d+)(,\d+)*\]", par_name, maxsplit=1)[0]
            if base_name == par:
                inds.append(k)
        except AttributeError:
            if par_name == par:
                inds.append(k)

    if len(inds) == 1:
        return data[:, :, inds[0]]

    return data[:, :, inds]


def hpd(x, mass_frac):
    """
    Returns highest probability density region given by
    a set of samples.

    Parameters
    ----------
    x : array
        1D array of MCMC samples for a single variable
    mass_frac : float with 0 < mass_frac <= 1
        The fraction of the probability to be included in
        the HPD.  For example, `massfrac` = 0.95 gives a
        95% HPD.

    Returns
    -------
    output : array, shape (2,)
        The bounds of the HPD
    """
    # Get sorted list
    d = np.sort(np.copy(x))

    # Number of total samples taken
    n = len(x)

    # Get number of samples that should be included in HPD
    n_samples = np.floor(mass_frac * n).astype(int)

    # Get width (in units of data) of all intervals with n_samples samples
    int_width = d[n_samples:] - d[: n - n_samples]

    # Pick out minimal interval
    min_int = np.argmin(int_width)

    # Return interval
    return np.array([d[min_int], d[min_int + n_samples]])

def posterior_predictive_ranking(fit=None, ppc_name=None, data=None, fractional=False):
    """Compute posterior predictive ranking of each data point.

    Parameters
    ----------
    fit : StanFit4Model instance
        A Stan fit of a model that has posterior predictive check values
        for the data.
    ppc_name : str
        The name of the parameter in the `fit` that has the posterior
        predictive checks.
    data : 1D Numpy array
        The data points that were observed.

    Returns
    -------
    output : 1D Numpy array, same shape as `data`
        The ranking among the posterior predictive samples of each data
        points.
    """
    if fit is None or ppc_name is None or data is None:
        raise RuntimeError("`fit`, `ppc_name`, and `data` must all be specified.")

    ppc = fit.extract(ppc_name)[ppc_name]
    rankings = np.empty(ppc.shape[1])

    for i in range(ppc.shape[1]):
        rankings[i] = np.searchsorted(np.sort(ppc[:, i]), data[i])

    if fractional:
        rankings /= ppc.shape[0]

    return rankings


def _infer_max_treedepth(fit):
    """Extract max treedepth from a StanFit4Model instance."""
    try:
        return fit.stan_args[0]["ctrl"]["sampling"]["max_treedepth"]
    except:
        warnings.warn("Unable to determine max_treedepth. " + "Using value of 10.")
        return 10
        

def _fit_to_az(fit, **kwargs):
    """Convert StanFit4Model to data frame."""
    if "StanFit4Model" in str(type(fit)):
        df = to_dataframe(fit, inc_warmup=False, **kwargs)
    elif type(fit) != pd.core.frame.DataFrame:
        raise RuntimeError("`fit` must be a StanModel or Pandas data frame.")
    else:
        df = fit.loc[fit["warmup"] == 0, :]

    return df



import collections
import warnings

import numpy as np
import pandas as pd

import emcee
import ptemcee

def generic_log_posterior(log_prior, log_likelihood, params, logpargs=(),
                          loglargs=()):
    """
    Generic log posterior for MCMC calculations

    Parameters
    ----------
    log_prior : function
        Function to compute the log prior.
        Call signature: log_prior(params, *logpargs)
    log_likelihood : function
        Function to compute the log prior.
        Call signature: log_likelhood(params, *loglargs)
    params : ndarray
        Numpy array containing the parameters of the posterior.
    logpargs : tuple, default ()
        Tuple of parameters to be passed to log_prior.
    loglargs : tuple, default ()
        Tuple of parameters to be passed to log_likelihood.

    Returns
    -------
    output : float
        The logarithm of the posterior evaluated at `params`.
    """
    # Compute log prior
    lp = log_prior(params, *logpargs)

    # If log prior is -inf, return that
    if lp == -np.inf:
        return -np.inf

    # Compute and return posterior
    return lp + log_likelihood(params, *loglargs)


def sampler_to_dataframe(sampler, columns=None):
    """
    Convert output of an emcee sampler to a Pandas DataFrame.

    Parameters
    ----------
    sampler : emcee.EnsembleSampler or emcee.PTSampler instance
        Sampler instance form which MCMC has already been run.

    Returns
    -------
    output : DataFrame
        Pandas DataFrame containing the samples. Each column is
        a variable, except: 'lnprob' and 'chain' for an
        EnsembleSampler, and 'lnlike', 'lnprob', 'beta_ind',
        'beta', and 'chain' for a PTSampler. These contain obvious
        values.
    """
    invalid_column_names = ['lnprob', 'chain', 'lnlike', 'beta',
                            'beta_ind']
    if np.any([x in columns for x in invalid_column_names]):
            raise RuntimeError('You cannot name columns with any of these: '
                                    + '  '.join(invalid_column_names))

    if columns is None:
        columns = list(range(sampler.chain.shape[-1]))

    if isinstance(sampler, emcee.EnsembleSampler):
        n_walkers, n_steps, n_dim = sampler.chain.shape

        df = pd.DataFrame(data=sampler.flatchain, columns=columns)
        df['lnprob'] = sampler.flatlnprobability
        df['chain'] = np.concatenate([i * np.ones(n_steps, dtype=int)
                                                for i in range(n_walkers)])
    elif isinstance(sampler, ptemcee.sampler.Sampler):
        n_temps, n_walkers, n_steps, n_dim = sampler.chain.shape

        df = pd.DataFrame(
            data=sampler.flatchain.reshape(
                (n_temps * n_walkers * n_steps, n_dim)),
            columns=columns)
        df['lnlike'] = sampler.loglikelihood.flatten()
        df['lnprob'] = sampler.logprobability.flatten()

        beta_inds = [i * np.ones(n_steps * n_walkers, dtype=int)
                     for i, _ in enumerate(sampler.betas)]
        df['beta_ind'] = np.concatenate(beta_inds)

        df['beta'] = sampler.betas[df['beta_ind']]

        chain_inds = [j * np.ones(n_steps, dtype=int)
                      for i, _ in enumerate(sampler.betas)
                      for j in range(n_walkers)]
        df['chain'] = np.concatenate(chain_inds)
    else:
        raise RuntimeError('Invalid sample input.')

    return df


def run_ensemble_emcee(log_post=None, n_burn=100, n_steps=100,
                       n_walkers=None, p_dict=None, p0=None, columns=None,
                       args=(), threads=None, thin=1, return_sampler=False,
                       return_pos=False):
    """
    Run emcee.

    Parameters
    ----------
    log_post : function
        The function that computes the log posterior.  Must be of
        the form log_post(p, *args), where p is a NumPy array of
        parameters that are sampled by the MCMC sampler.
    n_burn : int, default 100
        Number of burn steps
    n_steps : int, default 100
        Number of MCMC samples to take
    n_walkers : int
        Number of walkers, ignored if p0 is None
    p_dict : collections.OrderedDict
        Each entry is a tuple with the function used to generate
        starting points for the parameter and the arguments for
        the function.  The starting point function must have the
        call signature f(*args_for_function, n_walkers).  Ignored
        if p0 is not None.
    p0 : array
        n_walkers by n_dim array of initial starting values.
        p0[i,j] is the starting point for walk i along variable j.
        If provided, p_dict is ignored.
    columns : list of strings
        Name of parameters.  These will be the column headings in the
        returned DataFrame.  If None, either inferred from p_dict or
        assigned sequential integers.
    args : tuple
        Arguments passed to log_post
    threads : int
        Number of cores to use in calculation
    thin : int
        The number of iterations to perform between saving the
        state to the internal chain.
    return_sampler : bool, default False
        If True, return sampler as well as DataFrame with results.
    return_pos : bool, default False
        If True, additionally return position of the sampler.

    Returns
    -------
    df : pandas.DataFrame
        First columns give flattened MCMC chains, with columns
        named with the variable being sampled as a string.
        Other columns are:
          'chain':    ID of chain
          'lnprob':   Log posterior probability
    sampler : emcee.EnsembleSampler instance, optional
        The sampler instance.
    pos : ndarray, shape (nwalkers, ndim), optional
        Last position of the walkers.
    """

    if p0 is None and p_dict is None:
        raise RuntimeError('Must supply either p0 or p_dict.')

    # Infer n_dim and n_walkers (and check inputs)
    if p0 is None:
        if n_walkers is None:
            raise RuntimeError('n_walkers must be specified if p0 is None')

        if type(p_dict) is not collections.OrderedDict:
            raise RuntimeError('p_dict must be collections.OrderedDict.')

        n_dim = len(p_dict)
    else:
        n_walkers, n_dim = p0.shape
        if p_dict is not None:
            warnings.RuntimeWarning('p_dict is being ignored.')

    # Infer columns
    if columns is None:
        if p_dict is not None:
            columns = list(p_dict.keys())
        else:
            columns = list(range(n_dim))
    elif len(columns) != n_dim:
        raise RuntimeError('len(columns) must equal number of parameters.')

    # Check for invalid column names
    invalid_column_names = ['lnprob', 'chain', 'lnlike', 'beta',
                            'beta_ind']
    if np.any([x in columns for x in invalid_column_names]):
            raise RuntimeError('You cannot name columns with any of these: '
                                    + '  '.join(invalid_column_names))

    # Build starting points of walkers
    if p0 is None:
        p0 = np.empty((n_walkers, n_dim))
        for i, key in enumerate(p_dict):
            p0[:, i] = p_dict[key][0](*(p_dict[key][1] + (n_walkers,)))

    # Set up the EnsembleSampler instance
    if threads is not None:
        sampler = emcee.EnsembleSampler(n_walkers, n_dim, log_post,
                                        args=args, threads=threads)
    else:
        sampler = emcee.EnsembleSampler(n_walkers, n_dim, log_post,
                                        args=args)

    # Do burn-in
    if n_burn > 0:
        pos, _, _ = sampler.run_mcmc(p0, n_burn, storechain=False)
    else:
        pos = p0

    # Sample again, starting from end burn-in state
    pos, _, _ = sampler.run_mcmc(pos, n_steps, thin=thin)

    # Make DataFrame for results
    df = sampler_to_dataframe(sampler, columns=columns)

    # Set up return
    return_vals = (df, sampler, pos)
    return_bool = (True, return_sampler, return_pos)
    ret = tuple([rv for rv, rb in zip(return_vals, return_bool) if rb])
    if len(ret) == 1:
        return ret[0]
    return ret


def run_pt_emcee(log_like, log_prior, n_burn, n_steps, n_temps=None,
                 n_walkers=None, p_dict=None, p0=None, columns=None,
                 loglargs=(), logpargs=(), threads=None, thin=1,
                 return_lnZ=False, return_sampler=False, return_pos=False):
    """
    Run emcee.

    Parameters
    ----------
    log_like : function
        The function that computes the log likelihood.  Must be of
        the form log_like(p, *llargs), where p is a NumPy array of
        parameters that are sampled by the MCMC sampler.
    log_prior : function
        The function that computes the log prior.  Must be of
        the form log_post(p, *lpargs), where p is a NumPy array of
        parameters that are sampled by the MCMC sampler.
    n_burn : int
        Number of burn steps
    n_steps : int
        Number of MCMC samples to take
    n_temps : int
        The number of temperatures to use in PT sampling.
    n_walkers : int
        Number of walkers
    p_dict : collections.OrderedDict
        Each entry is a tuple with the function used to generate
        starting points for the parameter and the arguments for
        the function.  The starting point function must have the
        call signature f(*args_for_function, n_walkers).  Ignored
        if p0 is not None.
    p0 : array
        n_walkers by n_dim array of initial starting values.
        p0[k,i,j] is the starting point for walk i along variable j
        for temperature k.  If provided, p_dict is ignored.
    columns : list of strings
        Name of parameters.  These will be the column headings in the
        returned DataFrame.  If None, either inferred from p_dict or
        assigned sequential integers.
    args : tuple
        Arguments passed to log_post
    threads : int
        Number of cores to use in calculation
    thin : int
        The number of iterations to perform between saving the
        state to the internal chain.
    return_lnZ : bool, default False
        If True, additionally return lnZ and dlnZ.
    return_sampler : bool, default False
        If True, additionally return sampler.
    return_pos : bool, default False
        If True, additionally return position of the sampler.

    Returns
    -------
    df : pandas.DataFrame
        First columns give flattened MCMC chains, with columns
        named with the variable being sampled as a string.
        Other columns are:
          'chain':    ID of chain
          'beta':     Inverse temperature
          'beta_ind': Index of beta in list of betas
          'lnlike':   Log likelihood
          'lnprob':   Log posterior probability (with beta multiplying
                      log likelihood)
    lnZ : float, optional
        ln Z(1), which is equal to the evidence of the
        parameter estimation problem.
    dlnZ : float, optional
        The estimated error in the lnZ calculation.
    sampler : emcee.PTSampler instance, optional
        The sampler instance.
    pos : ndarray, shape (ntemps, nwalkers, ndim), optional
        Last position of the walkers.
    """

    if p0 is None and p_dict is None:
        raise RuntimeError('Must supply either p0 or p_dict.')

    # Infer n_dim and n_walkers (and check inputs)
    if p0 is None:
        if n_walkers is None:
            raise RuntimeError('n_walkers must be specified if p0 is None')

        if type(p_dict) is not collections.OrderedDict:
            raise RuntimeError('p_dict must be collections.OrderedDict.')

        n_dim = len(p_dict)
    else:
        n_temps, n_walkers, n_dim = p0.shape
        if p_dict is not None:
            warnings.RuntimeWarning('p_dict is being ignored.')

    # Infer columns
    if columns is None:
        if p_dict is not None:
            columns = list(p_dict.keys())
        else:
            columns = list(range(n_dim))
    elif len(columns) != n_dim:
        raise RuntimeError('len(columns) must equal number of parameters.')

    # Check for invalid column names
    invalid_column_names = ['lnprob', 'chain', 'lnlike', 'beta',
                            'beta_ind']
    if np.any([x in columns for x in invalid_column_names]):
            raise RuntimeError('You cannot name columns with any of these: '
                                    + '  '.join(invalid_column_names))

    # Build starting points of walkers
    if p0 is None:
        p0 = np.empty((n_temps, n_walkers, n_dim))
        for i, key in enumerate(p_dict):
            p0[:, :, i] = p_dict[key][0](
                *(p_dict[key][1] + ((n_temps, n_walkers),)))

    # Set up the PTSampler instance
    if threads is not None:
        sampler = ptemcee.Sampler(n_walkers, n_dim, log_like, log_prior,
                                  ntemps=n_temps, loglargs=loglargs,
                                  logpargs=logpargs, threads=threads)
    else:
        sampler = ptemcee.Sampler(n_walkers, n_dim, log_like, log_prior,
                                  ntemps=n_temps, loglargs=loglargs,
                                  logpargs=logpargs)

    # Do burn-in
    if n_burn > 0:
        pos, _, _ = sampler.run_mcmc(p0, iterations=n_burn, storechain=False)
    else:
        pos = p0

    # Sample again, starting from end burn-in state
    pos, _, _ = sampler.run_mcmc(pos, iterations=n_steps, thin=thin)

    # Compute thermodynamic integral
    lnZ, dlnZ = sampler.log_evidence_estimate(fburnin=0)

    # Make DataFrame for results
    df = sampler_to_dataframe(sampler, columns=columns)

    # Set up return
    return_vals = (df, lnZ, dlnZ, sampler, pos)
    return_bool = (True, return_lnZ, return_lnZ, return_sampler, return_pos)
    ret = tuple([rv for rv, rb in zip(return_vals, return_bool) if rb])
    if len(ret) == 1:
        return ret[0]
    return ret


def lnZ(df_mcmc):
    """
    Compute log Z(1) from PTMCMC traces stored in DataFrame.

    Parameters
    ----------
    df_mcmc : pandas DataFrame, as outputted from run_ptmcmc.
        DataFrame containing output of a parallel tempering MCMC
        run. Only need to contain columns pertinent to computing
        ln Z, which are 'beta_int', 'lnlike', and 'beta'.

    Returns
    -------
    output : float
        ln Z as computed by thermodynamic integration. This is
        equivalent to what is obtained by calling
        `sampler.thermodynamic_integration_log_evidence(fburnin=0)`
        where `sampler` is an emcee.PTSampler instance.

    Notes
    -----
    .. This is useful when the DataFrame from a PTSampler is too
       large to store in RAM.
    """
    # Average the log likelihood over the samples
    log_mean = np.zeros(len(df_mcmc['beta_ind'].unique()))
    for i, b in enumerate(df_mcmc['beta_ind'].unique()):
        log_mean[i] = df_mcmc['lnlike'][df_mcmc['beta_ind']==b].mean()

    # Set of betas (temperatures)
    betas = np.concatenate((np.array(df_mcmc['beta'].unique()), (0,)))

    # Approximate quadrature
    return np.dot(log_mean, -np.diff(betas))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))
import recommonmark.parser

# -- JUPYTER theme -----------------------------------------------------------
# from jupyter_sphinx_theme import *
# init_theme()
# globaltoc_depth = -1

# -- Project information -----------------------------------------------------

project = 'bebi103'
copyright = '2023, Justin Bois'
author = 'Justin Bois'

# The full version, including alpha/beta/rc tags
release = ''


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'recommonmark',
    'sphinx.ext.todo',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autosummary',
    'nbsphinx',
    'bokeh.sphinxext.bokeh_plot',
    'sphinx_copybutton',
    'sphinx.ext.napoleon'
]

autodoc_default_options = {
    'members': True,
    'inherited-members': True
}
autodoc_member_order = 'groupwise'

autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '**_build',
    '**Thumbs.db',
    '**.DS_Store',
    '**.ipynb_checkpoints',
    '**/__pycache__/**',
    '*/data/*',
    'Untitled**',
    '**.key',
    '**/.pytest_cache/**',
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# # This is processed by Jinja2 and inserted before each notebook
# nbsphinx_prolog = r"""
# {% set docname = env.doc2path(env.docname, base='doc') %}
# .. only:: html
#     .. role:: raw-html(raw)
#         :format: html
#     .. nbinfo::
#         This page was generated from `{{ docname }}`__.
#         Interactive online version:
#         :raw-html:`<a href="https://mybinder.org/v2/gh/spatialaudio/nbsphinx/{{ env.config.release }}?filepath={{ docname }}"><img alt="Binder badge" src="https://mybinder.org/badge_logo.svg" style="vertical-align:text-bottom"></a>`
#     __ https://github.com/spatialaudio/nbsphinx/blob/
#         {{ env.config.release }}/{{ docname }}
# .. raw:: latex
#     \nbsphinxstartnotebook{\scriptsize\noindent\strut
#     \textcolor{gray}{The following section was generated from
#     \sphinxcode{\sphinxupquote{\strut {{ docname | escape_latex }}}} \dotfill}}
# """

# # This is processed by Jinja2 and inserted after each notebook
# nbsphinx_epilog = r"""
# .. raw:: latex
#     \nbsphinxstopnotebook{\scriptsize\noindent\strut
#     \textcolor{gray}{\dotfill\ \sphinxcode{\sphinxupquote{\strut
#     {{ env.doc2path(env.docname, base='doc') | escape_latex }}}} ends here.}}
# """


# -- Options for HTML output -------------------------------------------------
# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

html_last_updated_fmt = ''

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

def setup(app):
    app.add_css_file('custom.css')

# html_logo = '_static/logo.png'

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}
#---sphinx-themes-----
html_theme = 'sphinx_rtd_theme'

html_theme_options = {
#    'analytics_id': 'UA-65442910-1',  #  Provided by Google in your dashboard
    'display_version': False,
    'logo_only': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'style_nav_header_background': '#F0F0F0',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
}



import warnings

import joblib

import numpy as np
import pandas as pd

import pymc3 as pm
import pymc3.stats
import pymc3.model
import theano.tensor as tt
import tqdm

from .hotdists import *

def _log_like_trace(trace, model, progressbar=False):
    """Calculate the elementwise log-likelihood for the sampled trace.
    Parameters
    ----------
    trace : result of MCMC run
    model : PyMC Model
        Optional model. Default None, taken from context.
    progressbar: bool
        Whether or not to display a progress bar in the command line. The
        bar shows the percentage of completion, the evaluation speed, and
        the estimated time to completion
    Returns
    -------
    logp : array of shape (n_samples, n_observations)
        The contribution of the observations to the log likelihood of 
        the whole model.

    Notes
    -----
    .. This is a copy of the pymc3.stats._log_post_trace() function for
       PyMC3 version 3.2. That is a misnomer, since it is not the log
       posterior for the trace, but rather the contributions to the 
       log posterior of each observation in the likelihood.
    """
    cached = [(var, var.logp_elemwise) for var in model.observed_RVs]

    def logp_vals_point(pt):
        if len(model.observed_RVs) == 0:
            return floatX(np.array([], dtype='d'))

        logp_vals = []
        for var, logp in cached:
            logp = logp(pt)
            if var.missing_values:
                logp = logp[~var.observations.mask]
            logp_vals.append(logp.ravel())

        return np.concatenate(logp_vals)

    try:
        points = trace.points()
    except AttributeError:
        points = trace

    points = tqdm.tqdm(points) if progressbar else points

    try:
        logp = (logp_vals_point(pt) for pt in points)
        return np.stack(logp)
    finally:
        if progressbar:
            points.close()


def _log_prior_trace(trace, model):
    """Calculate the elementwise log-prior for the sampled trace.
    Parameters
    ----------
    trace : result of MCMC run
    model : PyMC Model
        Optional model. Default None, taken from context.

    Returns
    -------
    logp : array of shape (n_samples, n_observations)
        The contribution of the log prior.
    """
    cached = [var.logp for var in model.unobserved_RVs 
                                if type(var) == pymc3.model.FreeRV]

    def logp_vals_point(pt):
        if len(model.unobserved_RVs) == 0:
            return floatX(np.array([], dtype='d'))

        return np.array([logp(pt) for logp in cached])

    try:
        points = trace.points()
    except AttributeError:
        points = trace

    logp = (logp_vals_point(pt) for pt in points)
    return np.stack(logp)


def _log_posterior_trace(trace, model):
    """
    Log posterior of each point in a trace.
    """
    return (_log_like_trace(trace, model).sum(axis=1) 
            + _log_prior_trace(trace, model).sum(axis=1))


def trace_to_dataframe(trace, model=None, log_post=False):
    """
    Convert a PyMC3 trace to a Pandas DataFrame

    Parameters
    ----------
    trace : PyMC3 trace
        Trace returned from pm.sample()
    model : PyMC3 model, default None
        Model returned from pm.Model()
    log_post : bool, default False
        If True, also compute the log posterior.

    Returns
    -------
    output : Pandas DataFrame
        DataFrame with samples and various sampling statistics.
    """
    df = pm.trace_to_dataframe(trace, chains=[0])
    for stat in trace.stat_names:
        if stat in df.columns:
            warnings.warn('`' + stat + '` is in the variable names.`'
                          + ' Not adding this statistic.')
        else:
            df[stat] = trace.get_sampler_stats(stat, chains=[0])
    if 'chain' in df.columns:
        warnings.warn('`chain` is in the variable name.`'
                          + ' Not adding this statistic.')
    else:
        df['chain'] = np.array([0]*len(df), dtype=int)

    if trace.nchains > 1:
        for chain in trace.chains[1:]:
            df_app = pm.trace_to_dataframe(trace, chains=[chain])
            for stat in trace.stat_names:
                if stat not in df_app.columns:
                    df_app[stat] = trace.get_sampler_stats(stat,
                                                           chains=[chain])
            if 'chain' not in df_app.columns:
                df_app['chain'] = np.array([chain]*len(df_app))

            df = df.append(df_app, ignore_index=True)

    if log_post:
        # Extract the model from context if necessary
        model = pm.modelcontext(model)

        df['log_likelihood'] = _log_like_trace(trace, model).sum(axis=1)
        df['log_prior'] = _log_prior_trace(trace, model).sum(axis=1)
        df['log_posterior'] = df['log_likelihood'] + df['log_prior']

    return df


class Jeffreys(pm.Continuous):
    """
    Jeffreys prior for a scale parameter.

    Parameters
    ----------
    lower : float, > 0
        Minimum value the variable can take.
    upper : float, > `lower`
        Maximum value the variable can take.

    Returns
    -------
    output : pymc3 distribution
        Distribution for Jeffreys prior.
    """
    def __init__(self, lower=None, upper=None, transform='interval',
                 *args, **kwargs):
        # Check inputs
        if lower is None or upper is None:
            raise RuntimeError('`lower` and `upper` must be provided.')

        if transform == 'interval':
            transform = pm.distributions.transforms.interval(lower, upper)
        super(Jeffreys, self).__init__(transform=transform, *args, **kwargs)        
        self.lower = lower = pm.theanof.floatX(tt.as_tensor_variable(lower))
        self.upper = upper = pm.theanof.floatX(tt.as_tensor_variable(upper))
        
        self.mean = (upper - lower) / tt.log(upper/lower)
        self.median = tt.sqrt(lower * upper)
        self.mode = lower
        
    def logp(self, value):
        lower = self.lower
        upper = self.upper
        return pm.distributions.dist_math.bound(
                    -tt.log(tt.log(upper/lower)) - tt.log(value),
                    value >= lower, value <= upper)


class MarginalizedHomoscedasticNormal(pm.Continuous):
    """
    Likelihood generated by marginalizing out a homoscedastic variance
    from a Normal distribution.

    Parameters
    ----------
    mu : array
        Mean of the distribution.

    Returns
    -------
    output : pymc3 distribution
        Distribution for a multivariate Gaussian with homoscedastic
        error, normalized over sigma.
    """
    def __init__(self, mu, *args, **kwargs):
        super(MarginalizedHomoscedasticNormal, self).__init__(*args, **kwargs)
        self.mu = mu = tt.as_tensor_variable(mu)
        self.mean = mu
        self.mode = mu


    def logp(self, value):
        n = value.shape[-1]
        prefactor = (  pm.distributions.dist_math.gammaln(n/2)
                     - tt.log(2)
                     - 0.5 * n * tt.log(np.pi)) 
        return prefactor - 0.5 * n * tt.log(tt.sum((value - self.mu)**2))


class GoodBad(pm.Continuous):
    """
    Likelihood for the good-bad data model, in which each data point
    is either "good" with a small variance or "bad" with a large 
    variance.

    Parameters
    ----------
    w : float
        Probability that a data point is "good."
    mu : float
        Mean of the distribution.
    sigma : float
        Standard deviation for "good" data points.
    sigma_bad : float
        Standard deviation for "bad" data points.

    Returns
    -------
    output : pymc3 distribution
        Distribution for the good-bad data model.
    """
    def __init__(self, mu, sigma, sigma_bad, w, *args, **kwargs):
        super(GoodBad, self).__init__(*args, **kwargs)
        self.mu = mu = tt.as_tensor_variable(mu)
        self.sigma = tt.as_tensor_variable(sigma)
        self.sigma_bad = tt.as_tensor_variable(sigma_bad)
        self.w = tt.as_tensor_variable(w)
        self.mean = mu
        self.median = mu
        self.mode = mu

    def logp(self, value):
        prefactor = -tt.log(2.0 * np.pi) / 2.0
        ll_good = (  tt.log(self.w / self.sigma) 
                   - ((value - self.mu) / self.sigma)**2 / 2.0)
        ll_bad = (  tt.log((1.0 - self.w) / self.sigma_bad)
                  - ((value - self.mu) / self.sigma_bad)**2 / 2.0)
        term = tt.switch(tt.gt(ll_good, ll_bad),
                         ll_good + tt.log(1 + tt.exp(ll_bad - ll_good)),
                         ll_bad + tt.log(1 + tt.exp(ll_good - ll_bad)))
        return prefactor + term
    

def ReparametrizedNormal(name, mu=None, sd=None, shape=1):
    """
    Create a reparametrized Normally distributed random variable.

    Parameters
    ----------
    name : str
        Name of the variable.
    mu : float
        Mean of Normal distribution.
    sd : float, > 0
        Standard deviation of Normal distribution.
    shape: int or tuple of ints, default 1
        Shape of array of variables. If 1, then a single scalar.

    Returns
    -------
    output : pymc3 distribution
        Distribution for a reparametrized Normal distribution.

    Notes
    -----
    .. The reparametrization procedure allows the sampler to sample
       a standard normal distribution, and then do a deterministic
       reparametrization to achieve sampling of the original desired 
       Normal distribution.
    """
    # Check inputs
    if type(name) != str:
        raise RuntimeError('`name` must be a string.')
    if mu is None or sd is None:
        raise RuntimeError('`mu` and `sd` must be provided.')

    var_reparam = pm.Normal(name + '_reparam', mu=0, sd=1, shape=shape)
    var = pm.Deterministic(name, mu + var_reparam * sd)

    return var


def ReparametrizedCauchy(name, alpha=None, beta=None, shape=1):
    """
    Create a reparametrized Cauchy distributed random variable.

    Parameters
    ----------
    name : str
        Name of the variable.
    alpha : float
        Mode of Cauchy distribution.
    beta : float, > 0
        Scale parameter of Cauchy distribution
    shape: int or tuple of ints, default 1
        Shape of array of variables. If 1, then a single scalar.

    Returns
    -------
    output : pymc3 distribution
        Reparametrized Cauchy distribution.

    Notes
    -----
    .. The reparametrization procedure allows the sampler to sample
       a Cauchy distribution with alpha = 0 and beta = 1, and then do a
       deterministic reparametrization to achieve sampling of the 
       original desired Cauchy distribution.
    """
    # Check inputs
    if type(name) != str:
        raise RuntimeError('`name` must be a string.')
    if alpha is None or beta is None:
        raise RuntimeError('`alpha` and `beta` must be provided.')

    var_reparam = pm.Cauchy(name + '_reparam', alpha=0, beta=1, shape=shape)
    var = pm.Deterministic(name, alpha + var_reparam * beta)

    return var


class Ordered(pm.distributions.transforms.ElemwiseTransform):
    """
    Class defining transform to order entries in an array.

    Code from Adrian Seyboldt from PyMC3 discourse: https://discourse.pymc.io/t/mixture-models-and-breaking-class-symmetry/208/4
    """
    name = 'ordered'

    def forward(self, x):
        out = tt.zeros(x.shape)
        out = tt.inc_subtensor(out[0], x[0])
        out = tt.inc_subtensor(out[1:], tt.log(x[1:] - x[:-1]))
        return out
    
    def forward_val(self, x, point=None):
        x, = pm.distributions.distribution.draw_values([x], point=point)
        return self.forward(x)

    def backward(self, y):
        out = tt.zeros(y.shape)
        out = tt.inc_subtensor(out[0], y[0])
        out = tt.inc_subtensor(out[1:], tt.exp(y[1:]))
        return tt.cumsum(out)

    def jacobian_det(self, y):
        return tt.sum(y[1:])


class Composed(pm.distributions.transforms.Transform):    
    """
    Class to build a transform out of an elementwise transform.

    Code from Adrian Seyboldt from PyMC3 discourse: https://discourse.pymc.io/t/mixture-models-and-breaking-class-symmetry/208/4   
    """
    def __init__(self, trafo1, trafo2):
        self._trafo1 = trafo1
        self._trafo2 = trafo2
        self.name = '_'.join([trafo1.name, trafo2.name])

    def forward(self, x):
        return self._trafo2.forward(self._trafo1.forward(x))
    
    def forward_val(self, x, point=None):
        return self.forward(x)

    def backward(self, y):
        return self._trafo1.backward(self._trafo2.backward(y))

    def jacobian_det(self, y):
        y2 = self._trafo2.backward(y)
        det1 = self._trafo1.jacobian_det(y2)
        det2 = self._trafo2.jacobian_det(y)
        return det1 + det2


def ordered_transform():
    """
    Make an ordered transform.

    Returns
    -------
    output : pm.distirbutions.transforms.Transform subclass instance
        Transform to order entries in tensor.

    Example
    -------
    To insist on ordering probabilities, p1 <= p2 <= p3,
    >>> p = pymc3.Beta('p',
                       alpha=1,
                       beta=1,
                       shape=3,
                       transform=ordered_transform())
    """
    return Composed(pm.distributions.transforms.LogOdds(), Ordered())


def hotdist(dist, name, beta_temp, *args, **kwargs):
    """
    Instantiate a "hot" distribution. The "hot" distribution takes the
    value returned by the logp method of `dist` and returns beta * logp.

    Parameters
    ----------
    dist : PyMC3 distribution
        The name of a distribution you want to make hot. Examples:
        pm.Normal, pm.Binomial, pm.MvNormal, pm.Dirichlet.
    name : str
        Name of the random variable.
    beta_temp : float on interval [0, 1]
        Beta value (inverse temperature) of the distribution.

    Returns
    -------
    output : pymc3 distribution
        Hot distribution. 
    """
    class HotDistribution(dist):
        def __init__(self, beta_temp, *args, **kwargs):
            super(HotDistribution, self).__init__(*args, **kwargs)
            if not (0 <= beta_temp <= 1):
                raise RuntimeError('Must have 0 ≤ beta_temp ≤ 1.')
            self.beta_temp = beta_temp

        def logp(self, value):
            return self.beta_temp * dist.logp(self, value)
        
    return HotDistribution(name, beta_temp, *args, **kwargs)


def beta_ladder(n=20, beta_min=1e-8):
    return np.logspace(np.log10(beta_min), 0, n)


def _sample_beta(beta, model_fun, args, kwargs):
    print(f'Sampling beta = {beta}....')

    model = model_fun(beta, *args)

    with model:
        trace = pm.sample(**kwargs)

    return trace, model, beta


def sample_beta_ladder(model_fun, betas, args=(), njobs=1, draws=500,
                       tune=500, progressbar=True, **kwargs):
    """
    Draw MCMC samples for a distribution for various values of beta.

    Parameters
    ----------
    model_fun : function
        Function that returns a PyMC3 model. Must have call signature
        model_fun(beta_temp, *args), where `beta_temp` is the inverse
        temperature.
    betas : array_like
        Array of beta values to sample.
    args : tuple, default ()
        Any additional arguments passed into `model_fun`.
    njobs : int, default 1
        Number of temperatures to run in parallel. This is *not* the
        number of samplers to run in parallel for each temperature.
        Each temperature is only sampled with a single walker.
    draws : int, default 500
        Number of samples to generate for each temperature.
    tune : int, default 500
        Number of tuning steps to take for each temperature.
    progressbar : bool, default True
        If True, show progress bars of samlers.
    kwargs 
        All additional kwargs are passed to pm.sample().

    Returns
    -------
    output : list of tuples
        List of tuples. The first tuple is the traces for each beta
        value. The second tuple is a tuple of compiled PyMC3 models.
        The last is a tuple of beta values.

    Example
    -------
    .. Draw samples out of a Normal distribution with a flat prior
       on `mu` and a HalfCauchy prior on `sigma`.

       x = np.random.normal(0, 1, size=100)

       def norm_model(beta_temp, beta_cauchy, x):
           with pm.Model() as model:
               mu = pm.Flat('mu')
               sigma = pm.HalfCauchy('sigma', beta=beta_cauchy)
               x_obs = HotNormal('x_obs', beta_temp=beta_temp, mu=mu, 
                                 sd=sigma, observed=x)
           return model

       betas = np.logspace(-3, 0, 10)
       tmb = sample_beta_ladder(norm_model, betas, args=(beta_cauchy, x))


    """
    # Insert code here to pop draws, tune, and progressbar out of kwargs

    if np.any(betas < 0) or np.any(betas > 1):
        raise RuntimeError('All beta values must be on interval (0, 1].')

    if not np.any(betas == 1):
        warnings.warn(
            'You probably want to sample beta = 1, the cold distribution.')

    if np.any(betas == 0):
        raise RuntimeError("Sampling beta = 0 not allowed;"
                           + " you're just sampling the prior in that case.")

    if len(betas) != len(np.unique(betas)):
        raise RuntimeError('Repeated beta entry.')

    kwargs['draws'] = draws
    kwargs['tune'] = tune
    kwargs['progressbar'] = progressbar

    if njobs == 1:
        return [_sample_beta(beta, model_fun, args, kwargs) for beta in betas]
    else:
        jobs = (joblib.delayed(_sample_beta)(beta, model_fun, args, kwargs)
                     for beta in betas)
        return joblib.Parallel(n_jobs=njobs)(jobs)


def log_evidence_estimate(trace_model_beta):
    """
    Compute an estimate of the log evidence.

    Parameters
    ----------
    trace_model_beta : list of (trace, model, beta) tuples
        List of (trace, model, beta) tuples as would be returned by
        sample_beta_ladder().

    Returns
    -------
    output : float
        Approximate negative log evidence.
    """

    # Extract traces, models, and betas
    betas = []
    traces = []
    models = []
    for tmb in trace_model_beta:
        traces.append(tmb[0])
        models.append(tmb[1])
        betas.append(tmb[2])

    betas = np.array(betas)

    if np.any(betas <= 0) or np.any(betas > 1):
        raise RuntimeError('All beta values must be between zero and one.')

    if len(betas) != len(np.unique(betas)):
        raise RuntimeError('Repeated beta entry.')

    # Sort
    inds = np.argsort(betas)
    betas = betas = [betas[i] for i in inds]
    traces =traces = [traces[i] for i in inds]
    models = models = [models[i] for i in inds]

    # Compute average log likelihood
    mean_log_like = []
    for beta, trace, model in zip(betas, traces, models):
        mean_log_like.append(_log_like_trace(trace, model).sum(axis=1).mean()
                                 / beta)

    # Add zero value
    betas = np.concatenate(((0,), betas))
    mean_log_like = np.concatenate(((mean_log_like[0],), mean_log_like))

    # Perform integral
    return np.trapz(mean_log_like, x=betas)


def chol_to_cov(chol, cov_prefix):
    """
    Convert flattened Cholesky matrix to covariance.

    Parameters
    ----------
    chol : array_like
        Lexicographically flattened Cholesky decomposition as returned
        from trace.get_values(chol), where trace is a PyMC3 MultiTrace
        instance.
    chol_prefix : str
        Prefix for the nam e of the covariance variable. Results are
        stored as prefix__i__j, where i and j are the row and column
        indices, respectively.

    Returns
    -------
    output : Pandas DataFrame
        DataFrame with values of samples of the components of the
        covariance matrix.
    """
    chol = np.array(chol)

    n = int(np.round((-1 + np.sqrt(8*chol.shape[1] + 1)) / 2))
    sigma = np.zeros_like(chol)
    inds = np.tril_indices(n)
    for i, r in enumerate(chol):
        L =  np.zeros((n, n))
        L[inds] = r
        sig = np.dot(L, L.T)
        sigma[i] = sig[inds]

    cols = ['{0:s}__{1:d}__{2:d}'.format(cov_prefix, i, j) 
                for i, j in zip(*inds)]
    return pd.DataFrame(columns=cols, data=sigma)


def hpd(x, mass_frac) :
    """
    Returns highest probability density region given by
    a set of samples.

    Parameters
    ----------
    x : array
        1D array of MCMC samples for a single variable
    mass_frac : float with 0 < mass_frac <= 1
        The fraction of the probability to be included in
        the HPD.  For example, `massfrac` = 0.95 gives a
        95% HPD.
        
    Returns
    -------
    output : array, shape (2,)
        The bounds of the HPD
    """
    # Get sorted list
    d = np.sort(np.copy(x))

    # Number of total samples taken
    n = len(x)
    
    # Get number of samples that should be included in HPD
    n_samples = np.floor(mass_frac * n).astype(int)
    
    # Get width (in units of data) of all intervals with n_samples samples
    int_width = d[n_samples:] - d[:n-n_samples]
    
    # Pick out minimal interval
    min_int = np.argmin(int_width)
    
    # Return interval
    return np.array([d[min_int], d[min_int+n_samples]])



# Import Bokeh modules for interactive plotting
import bokeh.plotting

# DataShader
import datashader as ds
import datashader.bokeh_ext as ds_bokeh_ext

# Display graphics in this notebook
bokeh.io.output_notebook()

def data_range(df, cols, margin=0.02):
    x_range = df[cols[0]].max() - df[cols[0]].min()
    y_range = df[cols[1]].max() - df[cols[1]].min()
    return ([df[cols[0]].min() - x_range*margin, df[cols[0]].max()+ - x_range*margin],
            [df[cols[1]].min() - y_range*margin, df[cols[1]].max()+ - y_range*margin])

x_range, y_range = data_range(df_mcmc, ['r1', 'p1'])

def create_image(x_range, y_range, w, h):
    cvs = ds.Canvas(x_range=x_range, y_range=y_range, plot_height=int(h),
                    plot_width=int(w))
    agg = cvs.points(df_mcmc, 'r1', 'p1')
    img = ds.transfer_functions.shade(agg, cmap=ds.colors.viridis,
                                      how='linear')
#    img = ds.transfer_functions.set_background(img, ds.colors.viridis[0])
    return ds.transfer_functions.dynspread(img, threshold=0.8)

p = bokeh.plotting.figure(height=350, width=400, x_range=x_range, y_range=y_range,
                          x_axis_label='r1', y_axis_label='p1')
ds_bokeh_ext.InteractiveImage(p, create_image)

ds.transfer_functions.shade?

def hist_xy(data, bins=100):
    h, b = np.histogram(data, bins=bins, density=True)
    x = np.empty(len(b) * 2)
    y = np.empty(len(b) * 2)
    x[::2] = b
    x[1::2] = b
    y[1:-1:2] = r1
    y[2::2] = r1
    y[0] = 0
    y[-1] = 0
    return x, y

def data_range(df, cols, margin=0.02):
    x_range = df[cols[0]].max() - df[cols[0]].min()
    y_range = df[cols[1]].max() - df[cols[1]].min()
    return ([df[cols[0]].min() - x_range*margin, df[cols[0]].max()+ - x_range*margin],
            [df[cols[1]].min() - y_range*margin, df[cols[1]].max()+ - y_range*margin])


def two_d_slice(df, cols):
    if len(cols) != 2:
        raise RuntimeError('cols must be list of length 2')

    x_range, y_range = data_range(df, cols)




n = 3
p1 = bokeh.plotting.figure(plot_height=200, plot_width=200)
p1.line(*hist_xy(df_mcmc['r1'], bins=100), color='black')
p2 = bokeh.plotting.figure(plot_height=200, plot_width=200)
p2.line(*hist_xy(df_mcmc['p1'], bins=100), color='black')
p3 = bokeh.plotting.figure(plot_height=200, plot_width=200,
                           x_range=x_range, y_range=y_range,
                           x_axis_label='r1', y_axis_label='p1')
p1.x_range = p3.x_range
p2.x_range = p3.y_range
p1.yaxis.visible = False
p1.xaxis.visible = False
p2.yaxis.visible = False

grid = bokeh.layouts.gridplot([[p1, None], [p3, p2]])
bokeh.io.show(grid)



import numpy as np
import numba
import scipy.special

_float_eps = 1.0e-14


def append_sort_index(Xstar, X, index_origin=0):
    """Add the entries of X to Xstar, sort the result, and find indices
    in the results where the entries in X appear. Typically, Xstar is
    an array of x-values for which a GP is to be coputed and X is an
    array of x-values for which there are data.

    Parameters
    ----------
    Xstar : array, shape (n,) or (n, d)
        Array whose unique, sorted entries are returned.
    X : array, shape (m,) or (m, d)
        Array to be added to Xstar and whose entries have their indices
        returned.
    index_origin : int, default 0
        Origin of indexing. Choose `index_origin=1` for use with Stan.

    Returns
    -------
    X_out : array, 1D if X is shape (n,), 2D if X is shape (n, d)
        Sorted, unique entries of concatenation of Xstar and X.
    inds : array, shape (n,)
        Array of indices such that X_out[inds] = X.
    """
    Xstar = np.array(Xstar)
    X = np.array(X)

    if len(X.shape) == 1:
        X_out = np.unique(np.concatenate((Xstar, X)))
        inds = np.searchsorted(X_out, X)
    else:
        if X.shape[1] != Xstar.shape[1]:
            raise RuntimeError("Shape mismatch in `Xstar` and `X`.")
        X_out = np.unique(np.concatenate((Xstar, X), axis=0), axis=0)
        inds = _searchsorted_vector(X_out, X)

    return X_out, inds + index_origin


@numba.njit
def _searchsorted_vector(X_unique, X):
    """Search a 2-D array that is sorted along the row axis.
    """
    X_unique_inds = np.empty(len(X), dtype=np.int64)

    for i, x in enumerate(X):
        j = 0
        while j < len(X_unique) and not (x == X_unique[j]).all():
            j += 1
        if j == len(X_unique):
            raise RuntimeError("Not finding.")
        X_unique_inds[i] = j

    return X_unique_inds

def cov_exp_quad(X1, X2=None, alpha=1.0, rho=1.0):
    """Return covariance matrix for squared exponential kernel.

    Parameters
    ----------
    X1 : 1D, shape (n,) or 2D array, shape (n, d)
        Array of n points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional.
    X2 : 1D, shape (m, ) or 2D array, shape (m, d) or None
        Array of m points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional. If None, assume m = 0 in output.
    alpha : float
        Marginalized standard deviation of the SE kernel.
    rho : float
        Length scale of the SE kernel.

    Returns
    -------
    output : array, shape(n + m, n + m)
    """
    if alpha <= 0:
        raise RuntimeError("`alpha` must be positive.")
    if rho <= 0:
        raise RuntimeError("`rho` must be positive.")

    X1 = _vector_to_array(X1)
    X2 = _vector_to_array(X2) if X2 is not None else None

    if X2 is None or (X1.shape == X2.shape and np.allclose(X1, X2)):
        return _se_covariance_matrix_sym(X1, alpha, rho)

    return _se_covariance_matrix(X1, X2, alpha, rho)


def cov_d1_exp_quad(x1, x2=None, alpha=1.0, rho=1.0):
    """Return covariance matrix for squared exponential kernel
    differentiated by the first variable.

    Parameters
    ----------
    x1 : array shape (n,)
        Array of n points to compute kernel.
    x2 : array shape (m,) or None
        Array of m points to compute kernel. If None, assume m = 0 in
        output.
    alpha : float
        Marginalized standard deviation of the SE kernel.
    rho : float
        Length scale of the SE kernel.

    Returns
    -------
    output : array, shape(n + m, n + m)
    """
    if alpha <= 0:
        raise RuntimeError("`alpha` must be positive.")
    if rho <= 0:
        raise RuntimeError("`rho` must be positive.")

    if x2 is None or (len(x1) == len(x2) and np.allclose(x1, x2)):
        return _d1_se_covariance_matrix_sym(x1, alpha, rho)

    return _d1_se_covariance_matrix(x1, x2, alpha, rho)


def cov_d1_d2_exp_quad(x, alpha=1.0, rho=1.0):
    """Return covariance matrix for squared exponential kernel
    differentiated once by the first variable and once by the second.

    Parameters
    ----------
    x : array shape (n,)
        Array of n points to compute kernel.
    alpha : float
        Marginalized standard deviation of the SE kernel.
    rho : float
        Length scale of the SE kernel.

    Returns
    -------
    output : array, shape(n + m, n + m)
    """
    if alpha <= 0:
        raise RuntimeError("`alpha` must be positive.")
    if rho <= 0:
        raise RuntimeError("`rho` must be positive.")

    return _d1_d2_se_covariance_matrix(x, alpha, rho)


def cov_matern(X1, X2=None, alpha=1.0, rho=1.0, nu=2.5):
    """Return covariance matrix for a Matérn kernel.

    Parameters
    ----------
    X1 : 1D, shape (n,) or 2D array, shape (n, d)
        Array of n points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional.
    X2 : 1D, shape (m, ) or 2D array, shape (m, d) or None
        Array of m points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional. If None, assume m = 0 in output.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.
    nu : float
        Smoothness parameter of the kernel.

    Returns
    -------
    output : array, shape(n + m, n + m)
    """
    if alpha <= 0:
        raise RuntimeError("`alpha` must be positive.")
    if rho <= 0:
        raise RuntimeError("`rho` must be positive.")
    if nu <= 0:
        raise RuntimeError("`nu` must be positive")

    X1 = _vector_to_array(X1)
    X2 = _vector_to_array(X2) if X2 is not None else None

    if X2 is None or (X1.shape == X2.shape and np.allclose(X1, X2)):
        if nu == 0.5:
            return _matern_1_covariance_matrix_sym(X1, alpha, rho)
        if nu == 1.5:
            return _matern_3_covariance_matrix_sym(X1, alpha, rho)
        if nu == 2.5:
            return _matern_5_covariance_matrix_sym(X1, alpha, rho)
        return _matern_covariance_matrix_sym(X1, alpha, rho, nu)

    if nu == 0.5:
        return _matern_1_covariance_matrix(X1, X2, alpha, rho)
    if nu == 1.5:
        return _matern_3_covariance_matrix(X1, X2, alpha, rho)
    if nu == 2.5:
        return _matern_5_covariance_matrix(X1, X2, alpha, rho)
    return _matern_covariance_matrix(X1, X2, alpha, rho, nu)


def cov_periodic(X1, X2=None, alpha=1.0, rho=1.0, T=1.0):
    """Return covariance matrix for a perdioic kernel.

    Parameters
    ----------
    X1 : 1D, shape (n,) or 2D array, shape (n, d)
        Array of n points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional.
    X2 : 1D, shape (m, ) or 2D array, shape (m, d) or None
        Array of m points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional. If None, assume m = 0 in output.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.
    T : float
        Period of the kernel.

    Returns
    -------
    output : array, shape(n + m, n + m)
    """
    if alpha <= 0:
        raise RuntimeError("`alpha` must be positive.")
    if rho <= 0:
        raise RuntimeError("`rho` must be positive.")
    if T <= 0:
        raise RuntimeError("`T` must be positive.")

    X1 = _vector_to_array(X1)
    X2 = _vector_to_array(X2) if X2 is not None else None

    if X2 is None or (X1.shape == X2.shape and np.allclose(X1, X2)):
        return _periodic_covariance_matrix_sym(X1, alpha, rho, T)

    return _periodic_covariance_matrix(X1, X2, alpha, rho, T)


def _vector_to_array(x):
    """Make sure input to kernel calculator two-dimensional."""
    x = np.array(x)

    if len(x.shape) == 1:
        return x.reshape((len(x), 1))

    return x


def cov_from_kernel(X1, X2, kernel, **kernel_params):
    """Return covariance matrix for specified kernel.

    Parameters
    ----------
    X1 : 1D, shape (n,) or 2D array, shape (n, d)
        Array of n points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional.
    X2 : 1D, shape (m, ) or 2D array, shape (m, d) or None
        Array of m points to compute kernel. If a 1D array, assume the
        points are one-dimensional. If a 2D array, assume the points are
        d-dimensional. If None, assume m = 0 in output.
    kernel : function
        Function of the form kernel(x1, x2, **kernel_params) that
        returns as scalar value that is the kernel evaluated at points
        x1, x2.
    """
    X1 = _vector_to_array(X1)
    X2 = _vector_to_array(X2) if X2 is not None else None

    if X2 is None or (X1.shape == X2.shape and np.allclose(X1, X2)):
        n = X1.shape[0]
        K = np.empty((n, n))

        for i in range(n):
            for j in range(i, n):
                K[i, j] = kernel(X1[i, :], X1[j, :], **kernel_params)
                K[j, i] = K[i, j]
    else:
        n, m = X1.shape[0], X2.shape[0]
        K = np.empty((n, m))

        for i in range(n):
            for j in range(m):
                K[i, j] = kernel(X1[i, :], X2[j, :], **kernel_params)

    return K


@numba.njit
def _se_covariance_matrix_sym(X, alpha, rho):
    """Return covariance matrix for squared exponential kernel."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _se_kernel(X[i, :], X[j, :], alpha, rho)
            K[j, i] = K[i, j]

    return K


def linear_kernel(x1, x2, sigma_b=0.0, sigma=1.0):
    """Linear kernel. k = sigma_b ** 2 + sigma ** 2 * np.dot(x1, x2)

    Parameters
    ----------
    x1 : float or array
        Point in the space of covariates.
    x2 : float or array
        Point in the space of covariates.
    sigma_b : float, default 0.0
        Offset of linear kernel.
    sigma : float or array, default 1.0
        scale of linear kernel. If an array, must be of the same length
        as x1 and x2.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is vectorial
    try:
        len_x1 = len(x1)
    except:
        x1 = np.array([x1])

    try:
        len_x2 = len(x2)
    except:
        x2 = np.array([x2])

    return sigma_b ** 2 + sigma ** 2 * np.dot(x1, x2)


def polynomial_kernel(x1, x2, sigma_b=1.0, sigma_p=1.0, d=1):
    """Polynomial kernel: (sigma_0^2 + sigma_p^2 x1 . x2)^d

    Parameters
    ----------
    x1 : float or array
        Point in the space of covariates.
    x2 : float or array
        Point in the space of covariates.
    sigma_b : float, default 1.0
        Offset of polynomial kernel.
    sigma_p : float, default 1.0
        Scale of polynomial kernel.
    d : int, default 1
        Degree of polynomial kernel

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is vectorial
    try:
        len_x1 = len(x1)
    except:
        x1 = np.array([x1])

    try:
        len_x2 = len(x2)
    except:
        x2 = np.array([x2])

    return (sigma_b ** 2 + sigma_p ** 2 * np.dot(x1, x2)) ** d


def se_kernel(x1, x2, alpha, rho):
    """Squared exponential kernel.

    Parameters
    ----------
    x1 : float or array
        Point in the space of covariates.
    x2 : float or array
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is vectorial
    try:
        len_x1 = len(x1)
    except:
        x1 = np.array([x1])

    try:
        len_x2 = len(x2)
    except:
        x2 = np.array([x2])

    return _se_kernel(x1, x2, alpha, rho)


def d1_se_kernel(x1, x2, alpha, rho):
    """Derivative of first variable of squared exponential kernel.

    Parameters
    ----------
    x1 : float
        Point in the space of covariates.
    x2 : float
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is float
    if not np.isscalar(x1) or not np.isscalar(x2):
        raise NotImplementedError(
            "Derivatives of kernels only allowed for scalar variables."
        )

    return _d1_se_kernel(x1, x2, alpha, rho)


def d2_se_kernel(x1, x2, alpha, rho):
    """Derivative of second variable of squared exponential kernel.

    Parameters
    ----------
    x1 : float
        Point in the space of covariates.
    x2 : float
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is float
    if not np.isscalar(x1) or not np.isscalar(x2):
        raise NotImplementedError(
            "Derivatives of kernels only allowed for scalar variables."
        )

    return _d2_se_kernel(x1, x2, alpha, rho)


def d1_d2_se_kernel(x1, x2, alpha, rho):
    """Mixed second derivative of squared exponential kernel.

    Parameters
    ----------
    x1 : float
        Point in the space of covariates.
    x2 : float
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is float
    if not np.isscalar(x1) or not np.isscalar(x2):
        raise NotImplementedError(
            "Derivatives of kernels only allowed for scalar variables."
        )

    return _d1_d2_se_kernel(x1, x2, alpha, rho)


def matern_kernel(x1, x2, alpha, rho, nu):
    """Matern kernel.

    Parameters
    ----------
    x1 : float or array
        Point in the space of covariates.
    x2 : float or array
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.
    nu : float
        Roughness parameter of the kernel
    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is vectorial
    try:
        len_x1 = len(x1)
    except:
        x1 = np.array([x1])

    try:
        len_x2 = len(x2)
    except:
        x2 = np.array([x2])

    if nu == 0.5:
        return _matern_1_kernel(x1, x2, alpha, rho)
    if nu == 1.5:
        return _matern_3_kernel(x1, x2, alpha, rho)
    if nu == 2.5:
        return _matern_5_kernel(x1, x2, alpha, rho)
    return _matern_kernel(x1, x2, alpha, rho, nu)


def periodic_kernel(x1, x2, alpha, rho, T):
    """Periodic kernel.

    Parameters
    ----------
    x1 : float or array
        Point in the space of covariates.
    x2 : float or array
        Point in the space of covariates.
    alpha : float
        Marginalized standard deviation of the kernel.
    rho : float
        Length scale of the kernel.
    T : float
        Period of kernel.

    Returns
    -------
    output : float
        Value of returned by kernel evaluated at x1, x2.

    """
    # Make sure input is vectorial
    try:
        len_x1 = len(x1)
    except:
        x1 = np.array([x1])

    try:
        len_x2 = len(x2)
    except:
        x2 = np.array([x2])

    return _periodic_kernel(x1, x2, alpha, rho, T)


@numba.njit
def _se_kernel(x1, x2, alpha, rho):
    """Squared exponential kernel."""
    x_diff = x1 - x2
    return alpha ** 2 * np.exp(-np.dot(x_diff, x_diff) / 2.0 / rho ** 2)


@numba.njit
def _d1_se_kernel(x1, x2, alpha, rho):
    """Derivative of first variable of squared exponential kernel."""
    x_diff = x1 - x2
    rho2 = rho ** 2

    return -(alpha ** 2) * x_diff * np.exp(-(x_diff ** 2) / 2.0 / rho2) / rho2


@numba.njit
def _d2_se_kernel(x1, x2, alpha, rho):
    """Derivative of first variable of squared exponential kernel."""
    return _d1_se_kernel(x2, x1, alpha, rho)


@numba.njit
def _d1_d2_se_kernel(x1, x2, alpha, rho):
    """Derivative of first variable of squared exponential kernel."""
    x_diff2 = (x1 - x2) ** 2
    rho2 = rho ** 2

    return (alpha / rho) ** 2 * np.exp(-x_diff2 / 2.0 / rho2) * (1 - x_diff2 / rho2)


def _matern_kernel(x1, x2, alpha, rho, nu):
    """Matern kernel."""
    x_diff = x1 - x2
    beta = np.sqrt(2 * nu * np.dot(x_diff, x_diff)) / rho

    # Special case where x1 = x2
    if beta == 0:
        return alpha ** 2

    ret_val = alpha ** 2 * 2 ** (1 - nu) * beta ** nu / scipy.special.gamma(nu)
    ret_val *= scipy.special.kv(nu, beta)
    return ret_val


@numba.njit
def _matern_1_kernel(x1, x2, alpha, rho):
    """Matern kernel with nu = 1/2."""
    x_diff = x1 - x2
    return alpha ** 2 * np.exp(-np.sqrt(np.dot(x_diff, x_diff)) / rho)


@numba.njit
def _matern_3_kernel(x1, x2, alpha, rho):
    """Matern kernel with nu = 3/2."""
    x_diff = x1 - x2
    x_diff2 = np.dot(x_diff, x_diff)

    exp_arg = np.sqrt(3.0 * x_diff2) / rho

    prefact = 1.0 + exp_arg

    return alpha ** 2 * prefact * np.exp(-exp_arg)


@numba.njit
def _matern_5_kernel(x1, x2, alpha, rho):
    """Matern kernel with nu = 5/2."""
    x_diff = x1 - x2
    x_diff2 = np.dot(x_diff, x_diff)

    exp_arg = np.sqrt(5.0 * x_diff2) / rho

    prefact = 1.0 + exp_arg + 5.0 * x_diff2 / 3.0 / rho ** 2

    return alpha ** 2 * prefact * np.exp(-exp_arg)


@numba.njit
def _periodic_kernel(x1, x2, alpha, rho, T):
    """Perdioic kernel."""
    x_diff = x1 - x2
    exp_arg = 2.0 / rho ** 2 * np.sin(np.pi / T * np.sqrt(np.dot(x_diff, x_diff))) ** 2

    return alpha ** 2 * np.exp(-exp_arg)


@numba.njit
def _se_covariance_matrix(X1, X2, alpha, rho):
    """Return covariance matrix for squared exponential kernel."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _se_kernel(X1[i, :], X2[j, :], alpha, rho)

    return K


@numba.njit
def _se_covariance_matrix_sym(X, alpha, rho):
    """Return covariance matrix for squared exponential kernel."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _se_kernel(X[i, :], X[j, :], alpha, rho)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _d1_se_covariance_matrix(x1, x2, alpha, rho):
    """Return covariance matrix for  derivative of first variable of
    squared exponential kernel."""
    n, m = len(x1), len(x2)
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _d1_se_kernel(x1[i], x2[j], alpha, rho)

    return K


@numba.njit
def _d1_se_covariance_matrix_sym(x, alpha, rho):
    """Return covariance matrix for  derivative of first variable of
    squared exponential kernel."""
    n = len(x)
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _d1_se_kernel(x[i], x[j], alpha, rho)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _d1_d2_se_covariance_matrix(x, alpha, rho):
    """Return covariance matrix for mixed second derivative of
    squared exponential kernel."""
    n = len(x)
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _d1_d2_se_kernel(x[i], x[j], alpha, rho)
            K[j, i] = K[i, j]

    return K


def _matern_covariance_matrix(X1, X2, alpha, rho, nu):
    """Return covariance matrix for Matérn kernel."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _matern_kernel(X1[i, :], X2[j, :], alpha, rho, nu)

    return K


def _matern_covariance_matrix_sym(X, alpha, rho, nu):
    """Return covariance matrix for Matérn kernel."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _matern_kernel(X[i, :], X[j, :], alpha, rho, nu)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _matern_1_covariance_matrix(X1, X2, alpha, rho):
    """Return covariance matrix for Matérn kernel with nu = 1/2."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _matern_1_kernel(X1[i, :], X2[j, :], alpha, rho)

    return K


@numba.njit
def _matern_1_covariance_matrix_sym(X, alpha, rho):
    """Return covariance matrix for Matérn kernel with nu = 1/2."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _matern_1_kernel(X[i, :], X[j, :], alpha, rho)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _matern_3_covariance_matrix(X1, X2, alpha, rho):
    """Return covariance matrix for Matérn kernel with nu = 3/2."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _matern_3_kernel(X1[i, :], X2[j, :], alpha, rho)

    return K


@numba.njit
def _matern_3_covariance_matrix_sym(X, alpha, rho):
    """Return covariance matrix for Matérn kernel with nu = 3/2."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _matern_3_kernel(X[i, :], X[j, :], alpha, rho)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _matern_5_covariance_matrix(X1, X2, alpha, rho):
    """Return covariance matrix for Matérn kernel with nu = 5/2."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _matern_5_kernel(X1[i, :], X2[j, :], alpha, rho)

    return K


@numba.njit
def _matern_5_covariance_matrix_sym(X, alpha, rho):
    """Return covariance matrix for Matérn kernel 5/2."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _matern_5_kernel(X[i, :], X[j, :], alpha, rho)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _periodic_covariance_matrix(X1, X2, alpha, rho, T):
    """Return covariance matrix for squared exponential kernel."""
    n, m = X1.shape[0], X2.shape[0]
    K = np.empty((n, m))

    for i in range(n):
        for j in range(m):
            K[i, j] = _periodic_kernel(X1[i, :], X2[j, :], alpha, rho, T)

    return K


@numba.njit
def _periodic_covariance_matrix_sym(X, alpha, rho, T):
    """Return covariance matrix for squared exponential kernel."""
    n = X.shape[0]
    K = np.empty((n, n))

    for i in range(n):
        for j in range(i, n):
            K[i, j] = _periodic_kernel(X[i, :], X[j, :], alpha, rho, T)
            K[j, i] = K[i, j]

    return K


@numba.njit
def _lower_tri_solve(L, b):
    """
    Solves the lower triangular system Lx = b.
    Uses column-based forward substitution, outlined in algorithm
    3.1.3 of Golub and van Loan.
    Parameters
    ----------
    L : ndarray
        Square lower triangulatar matrix (including diagonal)
    b : ndarray, shape L.shape[0]
        Right hand side of Lx = b equation being solved.
    Returns
    -------
    x : ndarray
        Solution to Lx = b.
    """
    n = L.shape[0]

    # Solve Lx = b.
    x = np.copy(b)
    for j in range(n - 1):
        if abs(L[j, j]) > _float_eps:
            x[j] /= L[j, j]
            for i in range(j + 1, n):
                x[i] -= x[j] * L[i, j]
        else:
            x[j] = 0.0

    if n > 0:
        if abs(L[n - 1, n - 1]) > _float_eps:
            x[n - 1] /= L[n - 1, n - 1]
        else:
            x[n - 1] = 0.0

    return x


@numba.njit
def _upper_tri_solve(U, b):
    """
    Solves the lower triangular system Ux = b.
    Uses column-based forward substitution, outlined in algorithm
    3.1.4 of Golub and van Loan.
    Parameters
    ----------
    U: ndarray
        Square upper triangulatar matrix (including diagonal)
    b : ndarray, shape L.shape[0]
        Right hand side of Ux = b equation being solved.
    Returns
    -------
    x : ndarray
        Solution to Ux = b.
    """
    n = U.shape[0]

    # Solve Ux = b by back substitution.
    x = np.copy(b)
    for j in range(n - 1, 0, -1):
        if abs(U[j, j]) > _float_eps:
            x[j] /= U[j, j]
            for i in range(0, j):
                x[i] -= x[j] * U[i, j]
        else:
            x[j] = 0.0

    if n > 0:
        if abs(U[0, 0]) > _float_eps:
            x[0] /= U[0, 0]
        else:
            x[0] = 0.0

    return x


def posterior_mean_cov(
    X,
    y,
    Xstar,
    sigma,
    kernel=se_kernel,
    include_deriv=False,
    delta=1e-8,
    **kernel_hyperparams
):
    """
    Compute the posterior mean vector and covariance matrix for a
    posterior Gaussian process derived from a Normal likelihood and
    Gaussian process prior.

    Parameters
    ----------
    X : 1D, shape (n,) or 2D array, shape (n, d)
        Array of n data points for which observations were made. If a 1D
        array, assume the points are one-dimensional. If a 2D array,
        assume the points are d-dimensional.
    y : array, shape (n,)
        Measured data points.
    Xstar : 1D, shape (nstar,) or 2D array, shape (nstar, d)
        Array of nstar data points for posterior predictions are to be
        made. If a 1D array, assume the points are one-dimensional. If a
        2D array, assume the points are d-dimensional.
    sigma : float or array, shape (n,)
        Standard deviation for Normal likelihood. If a float, assumed to
        be homoscedastic for all points.
    kernel : function, default se_kernel
        Kernel defining the Gaussian process. Must have call signature
        kernel(x1, x2, **kernel_hyperparams).
    include_deriv : bool, default False
        If True, include first derivatives in mean vectors and
        covariances. If True, `X` and `Xstar` must both be 1D because
        multivariate gradients are not implemented.
    delta : float, default 1e-8
        Small number, used to add to the diagonal of covariance matrices
        to ensure numerical positive definiteness.
    **kernel_hyperparams : kwargs
        All additional kwargs are sent as kwargs to the `kernel`
        function.

    Returns
    -------
    m : array, shape (nstar,)
        The mean function of the Gaussian process posterior evaluated at
        the points given by `Xstar`.
    Sigma : array, shape (nstar, nstar) or (2*nstar, 2*nstar)
        Covariance matrix of the Gaussian process posterior evaluated at
        the points given by `Xstar`.
    g : array, shape (nstar,)
        The derivative function of the Gaussian process posterior
        evaluated at the points given by `Xstar`. Only returned if
        `include_deriv` is True.
    Sigma_g : array, shape (nstar, nstar)
        The covariance matrix for the derivative of the Gaussian process
        posterior evaluated at the points given by `Xstar`. Only
        returned if `include_deriv` is True.

    Notes
    -----
    .. If include_deriv is True, X1 and Xstar must be 1D and a SE kernel
    must be used.
    """
    if include_deriv == 1:
        if len(X.shape) > 1 or len(Xstar.shape) > 1:
            raise NotImplementedError(
                "If `deriv` is True, then `X` and `Xstar` must be 1D."
            )
        if kernel != se_kernel:
            raise NotImplementedError(
                "If `deriv` is True, then `kernel` must be the default `se_kernel`."
            )

        alpha = kernel_hyperparams.get("alpha", 1.0)
        rho = kernel_hyperparams.get("rho", 1.0)

        return _posterior_mean_cov_deriv(X, y, Xstar, sigma, alpha, rho, delta)

    X = _vector_to_array(X)
    Xstar = _vector_to_array(Xstar)

    if np.isscalar(sigma):
        sigma2 = np.ones(len(X)) * sigma ** 2
    else:
        sigma2 = sigma ** 2

    # Build covariance matrices
    if kernel == se_kernel:
        Ky = cov_exp_quad(X, **kernel_hyperparams) + np.diag(sigma2)
        Kstar = cov_exp_quad(X, Xstar, **kernel_hyperparams)
        Kstarstar = cov_exp_quad(Xstar, Xstar, **kernel_hyperparams)
    elif kernel == matern_kernel:
        Ky = cov_matern(X, **kernel_hyperparams) + np.diag(sigma2)
        Kstar = cov_matern(X, Xstar, **kernel_hyperparams)
        Kstarstar = cov_matern(Xstar, Xstar, **kernel_hyperparams)
    elif kernel == periodic_kernel:
        Ky = cov_periodic(X, **kernel_hyperparams) + np.diag(sigma2)
        Kstar = cov_periodic(X, Xstar, **kernel_hyperparams)
        Kstarstar = cov_periodic(Xstar, Xstar, **kernel_hyperparams)
    else:
        Ky = cov_from_kernel(X, X, kernel, **kernel_hyperparams) + np.diag(
            sigma2
        )
        Kstar = cov_from_kernel(X, Xstar, kernel, **kernel_hyperparams)
        Kstarstar = cov_from_kernel(Xstar, Xstar, kernel, **kernel_hyperparams)

    return _solve_mean_cov(y, Ky, Kstar, Kstarstar, delta)


def _posterior_mean_cov_deriv(x, y, xstar, sigma, alpha, rho, delta):
    if np.isscalar(sigma):
        sigma2 = np.ones(len(x)) * sigma ** 2
    else:
        sigma2 = sigma ** 2

    Ky = cov_exp_quad(x, alpha=alpha, rho=rho) + np.diag(sigma2)
    Kstar = cov_exp_quad(x, xstar, alpha=alpha, rho=rho)
    Kstarstar = cov_exp_quad(xstar, xstar, alpha=alpha, rho=rho)

    d1_Kstar = cov_d1_exp_quad(x, xstar, alpha=alpha, rho=rho)
    d1_d2_Kstarstar = cov_d1_d2_exp_quad(xstar, alpha=alpha, rho=rho)

    # Solve for mstar and Sigmastar
    mstar, Sigmastar = _solve_mean_cov(y, Ky, Kstar, Kstarstar, delta)

    # Solve for gstar and Sigma_g_star
    neg_gstar, Sigma_g_star = _solve_mean_cov(y, Ky, d1_Kstar, d1_d2_Kstarstar, delta)

    return mstar, Sigmastar, -neg_gstar, Sigma_g_star


@numba.njit
def _solve_mean_cov(y, Ky, Kstar, Kstarstar, delta):
    # Solve for m_star
    L = np.linalg.cholesky(Ky)
    z = _lower_tri_solve(L, y)
    xi = _upper_tri_solve(L.transpose(), z)
    mstar = np.dot(Kstar.transpose(), xi)

    # Solve for Sigmastar
    Xi = np.empty_like(Kstar)
    for j in range(Xi.shape[1]):
        z = _lower_tri_solve(L, Kstar[:, j])
        Xi[:, j] = _upper_tri_solve(L.transpose(), z)
    Sigmastar = (
        Kstarstar
        - np.dot(Kstar.transpose(), Xi)
        + np.diag(np.ones(len(Kstarstar)) * delta)
    )

    return mstar, Sigmastar



# -*- coding: utf-8 -*-

"""Top-level package for bebi103."""

# Force showing deprecation warnings.
import re
import warnings

warnings.filterwarnings(
    "always", category=DeprecationWarning, module=r"^{}\.".format(re.escape(__name__))
)

# Filter annoying warning about PyArrow
warnings.filterwarnings('ignore', message='.*pyarrow.*', category=FutureWarning)

try:
    import multiprocess
except:
    warnings.warn(
        "Unable to import multiprocess. Using multiprocessing (note the"
        " ing) instead. Depending on your operating system, Python"
        " version, and whether or not you are running in Jupyter, "
        " IPython, the Python REPL, etc., your execution may stall if"
        " you try to run jobs on more than one core. See discussion"
        " here: https://github.com/ipython/ipython/issues/12396. As a"
        " workaround, you can install multiprocess"
        " (pip install multiprocess) and everything should work as"
        " expected.",
        ImportWarning,
    )

from . import hv

from . import viz

from . import image

from . import bootstrap

from . import gp

from . import utils

try:
    from . import stan
except:
    warnings.warn(
        "Could not import `stan` submodule. Perhaps ArviZ or PyStan or CmdStanPy is/are"
        " not properly installed."
    )


__author__ = """Justin Bois"""
__email__ = "bois@caltech.edu"
__version__ = "0.1.22"



from setuptools import setup, find_packages
from codecs import open
from os import path
import json

def read(fname):
    with open(fname, mode = "r", encoding = "utf-8") as f:
        src = f.read()
    return src

codemeta_json = "codemeta.json"

# Let's pickup as much metadata as we need from codemeta.json
with open(codemeta_json, mode = "r", encoding = "utf-8") as f:
    src = f.read()
    meta = json.loads(src)

# Let's make our symvar string
__version__ = meta["version"]

# Now we need to pull and format our author, author_email strings.
author = ""
author_email = ""
for obj in meta["author"]:
    given = obj["givenName"]
    family = obj["familyName"]
    email = obj["email"]
    if len(author) == 0:
        author = given + " " + family
    else:
        author = author + ", " + given + " " + family
    if len(author_email) == 0:
        author_email = email
    else:
        author_email = author_email + ", " + email
description = meta['description']
url = meta['codeRepository']
license = meta['license']
name = meta['name']
keywords= meta['keywords']

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if x.startswith('git+')]

setup(
    name=name,
    version=__version__,
    description=description,
    long_description=long_description,
    url=url,
    download_url='https://github.com/justinbois/bebi103/tarball/' + __version__,
    license='BSD',
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 3',
    ],
    keywords=keywords,
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    author=author,
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email=author_email
)
